# !/bin/ksh
# File Name:     STATSCHK.ksh
#
# Author:        Alex Levy t/a Sustainable Software Ltd. http://www.sustainablesoftware.net
#
# Synopsis:      This utility creates a list of tables to RUNSTATS. It is functionally equivalent to the lower case statschk.ksh, 
#                with all function calls replaced by call to a built-in function library, shipped with  
#                http://www.sustainablesoftware.net/downloads/V9funclib.tar.gz 
#
#                By default, it runs the db2pd utility and lists all tables in permanent tablespaces with a positive Update Delete Insert (UDI) counter.
#                You can configure the candidate list so that it only considers tables in an optional list of tablespaces.
#                To do this, add the tablespaceid of each such tablespace on a separate line in the statschk_search_tbspaces file.
#                To consider all permanent tablespaces (i.e those not used for temporary data), just leave that file empty.
#                You can also specify an optional list of tablespaces whose tables must always be RUNSTATSd, regardless of UDI counts.
#                To do this, add the tablespaceid of each such tablespace on a separate line in the statschk_include_tbspaces file.
#                The assumption here is that related objects are grouped in the same tablespaces.
#
#                There is a final option to invoke RUNSTATS for each identified table.
#                Set the command line argument $2 to STATS to request this feature, otherwise set to NOSTATS.
#                The program will use a statistics profile, if one exists.  
#                Review the RUNSTATS() function options for tables without an existing statistics profile, if you run with the STATS option.
#
# Notes:         By default the program will create log and temporary directories under the current directory.
#                Edit the values for LOGDIR and TMPDIR if you want somewhere else for generated log and temporary work files.
#                
#                See the aaa_readme file for a high level overview of options for identifying tables to RUNSTATS.
#
# File 
#  Dependencies: $INSTHOME/sqllib/db2profile (in)
#                $LOGDIR and $TMPDIR (in/out)
#                1.awkin and 2.awkin (in) shipped in the same directory as this program.
#                The two files statschk_search_tbspaces and statschk_include_tbspaces are shipped empty
#                in the same directory as this program.
#           
# Database
#  Dependencies: Activated database
#
# Parameters:    $1 = <source database name> e.g. sample
#                $2 = either STATS or NOSTATS, your choice whether you wish to execute RUNSTATS straightaway.
#                $3 = optional ksh debug switch, either -debug Y (debug) or -debug N (no debug) 
#                If $3 is not supplied, no debug is assumed. 
#                e.g. /home/dba/STATSCHK.ksh sample nostats -debug N. If $3 is not supplied, no debug is assumed. 
#
# Assumptions:   - environment variableis $DB2INSTANCE and $INSTHOME (set by $INSTHOME/sqllib/userprofile) are exposed
#       	   (this is part of the setup for the built-in function library).  
#       
# Date:          Jan 2010
# 
# AMENDMENT HISTORY
# =================
# 
# 
#---------------------------------------------------------------------------------------

# ----- Global variables -------------
BOLD=$(tput smso) ; PLAIN=$(tput rmso)
typeset -i RC=0
typeset -l debug="+x"
export FULL_SCRIPTNAME=$0
WORKDIR=$(dirname $0)	# current directory of this program
LOGDIR=${WORKDIR}/log	# log and temp directories are created further down, edit locations as required
TMPDIR=${WORKDIR}/tmp	#
logprefix=$(basename $0 .ksh)
TS=$(date +"%Y%m%d.%H%M%S")
export LOG=${LOGDIR}/${logprefix}.$TS 
# -----------------------------------

 
RESTORE_OUTPUT () {
 RESET_OUTPUT      # Call function to restore output back to stdin and stderr.
 if  [[ -s $LOG ]] # Display the log file to stdout.
 then
     cat $LOG
     print "\nLog is $LOG and is reproduced above.\n"
 fi
}
INITIAL () {
 export DEBUG_SWITCH=$debug
 if [[ $# -eq 2 || $# -eq 4 ]]
 then
    typeset -u dbname=$1
    typeset -u option=$2
    case $option in
         STATS|NOSTATS) 
               : # no action
               ;;
         *)    SYNTAX "dbname [STATS|NOSTATS]"
               ;; 
    esac 
    if [[ -n $4 ]]
    then 
       typeset -u $4
       case $4 in
            Y) 
            debug="-x"
               ;;
            N) 
            debug="+x"
               ;;
           *) SYNTAX "dbname [STATS|NOSTATS]"
               ;;
       esac
       export DEBUG_SWITCH=$debug
    fi   
 else 
    SYNTAX "dbname [STATS|NOSTATS]"
 fi

 # Create log and temporary work directories if they don't already exist
 if [[ -d $LOGDIR && -d $TMPDIR ]]
 then chmod 755 $LOGDIR $TMPDIR
 else
      mkdir -m 755 $LOGDIR $TMPDIR 2> /dev/null
 fi

 INIT_OUTPUT_FILE $LOG $debug # create and head log file;
 REDIRECT_OUTPUT  $LOG $debug # redirect stdout/stderr log messages to log file
 print "$0 $TS ===> Call is $0 $* "

 # test all required files are present and readable:
 EXISTENCE_TEST ${WORKDIR}/1.awkin $debug                    ; RC=$? ; RETURNCODE_CHECK $RC 0 $debug 
 EXISTENCE_TEST ${WORKDIR}/2.awkin $debug                    ; RC=$? ; RETURNCODE_CHECK $RC 0 $debug
 EXISTENCE_TEST ${WORKDIR}/statschk_search_tbspaces  $debug  ; RC=$? ; RETURNCODE_CHECK $RC 0 $debug
 EXISTENCE_TEST ${WORKDIR}/statschk_include_tbspaces $debug  ; RC=$? ; RETURNCODE_CHECK $RC 0 $debug
  
 SETUP_DB2_ENVIRONMENT $INSTHOME $DB2INSTANCE $dbname $debug # executes the instance owner's db2profile
 RC=$? ; RETURNCODE_CHECK $RC 0 $debug

 # test for an activated database as db2pd table statistics will be of little use otherwise; 
 # a first connection will activate an inactive database, so we have to check in a rather roundabout way;
 # first set a space-terminated variable for the database alias so a grep will not pick a false positive,
 # e.g., finding SAMPLE1 when it is looking for SAMPLE. 
 db=${dbname}" "
 typeset -i active=$(db2 "list active databases" | grep -c $db )
 if [[ $active -eq 1 ]]
 then
    : # no action
 else
    MESSAGE "$0 : database $dbname is not activated. Exiting ..." BOLD E $debug
    RESTORE_OUTPUT 
    exit 3
 fi
 
# Connection test - db2pd -tcbstats by itself does not require a connection, but catalog searches do.
 CONNECT $dbname $debug
 RC=$? ; RETURNCODE_CHECK $RC 0 $debug

}

MAIN () {

 typeset -u dbname=$1 runstats_option=$2
 # See whether the caller wants to limit the statistics check to just some tablespaces;
 # if not, select all regular and large tablespaces with permanent data (no temporary or user temporary tablespaces).
 if [[ -s ${WORKDIR}/statschk_search_tbspaces ]]
 then
    # file is populated, copy to temp work directory
    cp -p ${WORKDIR}/statschk_search_tbspaces ${TMPDIR}/statschk_search_tbspaces.${TS} 
 else
    db2 -x "select tbspaceid 	         \
	    from   syscat.tablespaces	 \
	    where  datatype in ('A','L') \
	    order by tbspaceid for fetch only" > ${TMPDIR}/statschk_search_tbspaces.${TS}
    RC=$? 
    if [[ $RC -gt 0 ]]
    then MESSAGE "$0 Query error on catalog view syscat.tablespaces. Detail in log above." BOLD E $debug
         RETURNCODE_CHECK $RC 0 $debug 
    fi
 fi

 # The awkin scripts comment the logic for parsing db2pd output.
 TOUCH_FILE ${TMPDIR}/1.awk.out.${TS} $debug
 TOUCH_FILE ${TMPDIR}/2.awk.out.${TS} $debug
 for tbspaceid in `cat ${TMPDIR}/statschk_search_tbspaces.${TS}`
 do
     eval db2pd -tcbstats TbspaceID=${tbspaceid} -db ${dbname} -o ${TMPDIR}/db2pd.out.${tbspaceid}.${TS} 1> /dev/null # suppress informational messages
     # this return code check is of limited use, since db2pd still produces RC=0 for bad arguments:
     RC=$? 
     if [[ $RC -gt 0 ]]
     then MESSAGE "$0 call error on db2pd -tcbstats TbspaceID=${tbspaceid} -db ${dbname} -o ${TMPDIR}/db2pd.out.${tbspaceid}.${TS}" BOLD E $debug
          RETURNCODE_CHECK $RC 0 $debug
     fi
     awk -f ${WORKDIR}/1.awkin ${TMPDIR}/db2pd.out.${tbspaceid}.${TS} >> ${TMPDIR}/1.awk.out.${TS}
     awk -f ${WORKDIR}/2.awkin ${TMPDIR}/db2pd.out.${tbspaceid}.${TS} >> ${TMPDIR}/2.awk.out.${TS}
 done

 # Sort both outputs on join field (table address)
 sort ${TMPDIR}/1.awk.out.${TS} > ${TMPDIR}/1.awk.out.sorted.${TS}
 sort ${TMPDIR}/2.awk.out.${TS} > ${TMPDIR}/2.awk.out.sorted.${TS}
 # Join on table address to obtain the initial list of fully qualified tables for RUNSTATS:
 join -o 1.2 1.3 -1 1  ${TMPDIR}/1.awk.out.sorted.${TS} ${TMPDIR}/2.awk.out.sorted.${TS} | sed 's/ /./' | sort -u > ${TMPDIR}/${logprefix}.tablist.${TS}
 
 # Add any tables that will always be RUNSTATS'D regardless
 if [[ -s ${WORKDIR}/statschk_include_tbspaces ]]
 then
    # file has size > 0
    for tbspaceid in `cat ${WORKDIR}/statschk_include_tbspaces`
    do
       db2 -x "select distinct rtrim(tabschema)||'.'||tabname \
               from   syscat.datapartitions                   \
               where  tbspaceid = $tbspaceid   		   \
               order  by 1 for fetch only" >> ${TMPDIR}/${logprefix}.tablist.${TS}
       RC=$? 
       if [[ $RC -gt 0 ]] 
       then MESSAGE "$0 Query error on catalog view syscat.datapartitions. Detail in log above." BOLD E $debug
            RETURNCODE_CHECK $RC 0
       fi
    done 
    # The search and include files could overlap, so remove duplicates:
    sort -u ${TMPDIR}/${logprefix}.tablist.${TS} > ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}
 else
    mv ${TMPDIR}/${logprefix}.tablist.${TS} ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}
 fi

 # RUNSTATS on identified tables, if requested:
 if [[ $runstats_option = "STATS" ]]
 then
    for name in `cat ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}`
    do
        schema=${name%%\.*}
        tabname=${name##*\.}
        RUNSTATS $schema $tabname $debug
        RC=$? ; RETURNCODE_CHECK $RC 0 $debug
    done
 fi
}

FINAL () {
 typeset -u option=$1
 TERMINATE $debug 
 RC=$? ; RETURNCODE_CHECK $RC 0 $debug
 # remove temporary work files from previous runs (those with a timestamp suffix not matching this run)
 find ${TMPDIR} \! -name "*.${TS}" -type f -exec rm -f {} \;
 if [[ $option = "STATS" ]]
 then
    print "$0 : RUNSTATS was executed for tables in list $LOGDIR}/${logprefix}.runstats_tablelist.${TS}"
 else
    print "$0 : The list of tables to consider for RUNSTATS is in ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}"
 fi
 print "$(date +"%d/%m/%Y %H:%M:%S") ===> $0 ends with maximum return code $RC."
 RESTORE_OUTPUT 
}

INITIAL $1 $2
MAIN    $1 $2 
FINAL   $2   
