# !/bin/ksh
# File Name:     statschk.ksh
#
# Author:        Alex Levy t/a Sustainable Software Ltd. http://www.sustainablesoftware.net
#
# Synopsis:      This utility creates a list of tables to RUNSTATS.
#                By default, it runs the db2pd utility and lists all tables in permanent tablespaces with a positive Update Delete Insert (UDI) counter.
#                You can configure the candidate list so that it only considers tables in an optional list of tablespaces.
#                To do this, add the tablespaceid of each such tablespace on a separate line in the statschk_search_tbspaces file.
#                To consider all permanent tablespaces (i.e those not used for temporary data), just leave that file empty.
#                You can also specify an optional list of tablespaces whose tables must always be RUNSTATSd, regardless of UDI counts.
#                To do this, add the tablespaceid of each such tablespace on a separate line in the statschk_include_tbspaces file.
#                The assumption here is that related objects are grouped in the same tablespaces.
#
#                There is a final option to invoke RUNSTATS for each identified table.
#                Set the command line argument $2 to STATS to request this feature, otherwise set to NOSTATS.
#                The program will use a statistics profile, if one exists.  
#                Review the RUNSTATS() function options for tables without an existing statistics profile, if you run with the STATS option.
#
# Notes:         By default the program will create log and temporary directories under the current directory.
#                Edit the values for LOGDIR and TMPDIR if you want somewhere else for generated log and temporary work files.
#                
#                See the aaa_readme file for a high level overview of options for identifying tables to RUNSTATS.
#
# File 
#  Dependencies: $INSTHOME/sqllib/db2profile (in)
#                $LOGDIR and $TMPDIR (in/out)
#                1.awkin and 2.awkin (in) shipped in the same directory as this program.
#                The two files statschk_search_tbspaces and statschk_include_tbspaces are shipped empty
#                in the same directory as this program.
#           
# Database
#  Dependencies: Activated database
#
# Parameters:    $1 = <source database name> e.g. sample
#                $2 = either STATS or NOSTATS, your choice whether you wish to execute RUNSTATS straightaway.
#
# Assumptions:   - Runs as the instance owner ;
#                - environment variable $INSTHOME (set by $INSTHOME/sqllib/db2profile) is exposed.
#
# Date:          Jan 2010
# 
# AMENDMENT HISTORY
# =================
# 
# 
#---------------------------------------------------------------------------------------

# -----Global variables
BOLD=$(tput smso) ; PLAIN=$(tput rmso)
typeset -i RC=0
WORKDIR=$(dirname $0)	# current directory of this program
LOGDIR=${WORKDIR}/log	# log and temp directories are created further down, edit locations as required
TMPDIR=${WORKDIR}/tmp	#
logprefix=$(basename $0 .ksh)
TS=$(date +"%Y%m%d.%H%M%S")
LOG=${LOGDIR}/${logprefix}.$TS 
# --------------------

USAGE () {
 print ${BOLD} "Usage!: I said $0 database execute_runstats_option"
 print              "       database: local database alias e.g. sample"
 print              "       execute_runstats_option: either STATS or NOSTATS" ${PLAIN}
 exit 1
}
 
RESTORE_OUTPUT () {
 # Restore output back to stdin and stderr.
 exec 1>&-
 exec 1>&2
 if  [[ -s $LOG ]]
 then 
     cat $LOG
     print "\nLog is $LOG and is reproduced above.\n"
 fi
}

DB2_RC_CHECK  () {
 typeset -i RETCODE=$1 MAXRC=$2 
 LOCATION=$3  
 MSG=$4
 if [[ $RETCODE -gt $MAXRC ]]
 then 
    print ${BOLD} $0 DB2/SQL error return code $RETCODE at program location $LOCATION.
    if [[ -n $MSG ]]
    then print $MSG  
    fi
    print Exiting ... ${PLAIN}
    db2 -v terminate
    RESTORE_OUTPUT
    exit $RETCODE
fi
}

RUNSTATS() {
# RUNSTATS a given schema/table combination.
# Uses a stats profile if one exists, otherwise sets the profile as well.
# Caller establishes the database connection.
# 
# Arguments $1 schema name - will fold to upper case; 
#           $2 table name  - unqualified 
# e.g. RUNSTATS $DBSCHEMA EMPLOYEE . 

 if [[ $# -eq 2 ]]
 then
    typeset -u SCHEMA=$1 TABLE=$2 
    TARGET=$SCHEMA.$TABLE
    print "Call is RUNSTATS $*"
 else
    print "RUNSTATS() Bad Call! \n Usage RUNSTATS \$1 schema name \$2 table name \n" 
    return 1 
 fi    
 
 typeset -i PROFILE=$(db2 -x "select count(*) 				\
                              from   syscat.tables 			\
                              where  tabschema = '${SCHEMA}' 		\
                              and    tabname = '${TABLE}' 		\
                              and    statistics_profile is not null 	\
                              and    type = 'T' 			\
                              for fetch only")
 RC=$? ; DB2_RC_CHECK $RC 1 1 "RUNSTATS() :Error querying SYSCAT.TABLES for stats profile for $TARGET. See stderr on $LOG ."  

 case $PROFILE in
      0) # runstats and set profile in one pass
         db2 -v "runstats on table $TARGET on all columns with distribution on all columns and detailed indexes all allow read access set profile"
         RC=$? ; DB2_RC_CHECK $RC 0 2 "RUNSTATS() :Error using default profile for $TARGET. See stderr on $LOG ."  
         ;;
      1) # stats profile exists, let's use it
         db2 -v "runstats on table $TARGET use profile" 
         RC=$? ; DB2_RC_CHECK $RC 0 3 "RUNSTATS() :Error using existing statistics profile for $TARGET. See stderr on $LOG ."  
         ;;
      *) print "RUNSTATS() Logic error in code, Alex please review" 
         return 1
         ;;
 esac
}

INITIAL () {
 if [[ $# -eq 2 ]]
 then
    typeset -u dbname=$1
    typeset -u option=$2
    case $option in
         STATS|NOSTATS) 
               # Re-direct stdin/stderr to log file.
               exec 2>&1
               exec 1>> $LOG
               print "$0 $TS ===> Call is $0 $* "
               ;;
         *)    USAGE
               ;; 
    esac 
 else 
    USAGE
 fi

 # Create log and temporary work directories if they don't already exist
 if [[ -d $LOGDIR && -d $TMPDIR ]]
 then chmod 755 $LOGDIR $TMPDIR
 else
      mkdir -m 755 $LOGDIR $TMPDIR 2> /dev/null
 fi
 
 # test all required files are present and readablei, and that awk scripts are populated:
 if [[ -s ${WORKDIR}/1.awkin && -s ${WORKDIR}/2.awkin && \
       -r ${WORKDIR}/1.awkin && -r ${WORKDIR}/2.awkin && \
       -f ${WORKDIR}/statschk_search_tbspaces && -f ${WORKDIR}/statschk_include_tbspaces && \
       -r ${WORKDIR}/statschk_search_tbspaces && -r ${WORKDIR}/statschk_include_tbspaces ]]   
 then 
    print "\n Log is $LOG \n"
 else
    print ${BOLD} "$0 : At lease one input file is missing or corrupt. Reinstall !" ${PLAIN}
    RESTORE_OUTPUT 
    exit 2
 fi
  
 # Set DB2 environment
 . $INSTHOME/sqllib/db2profile
 # test for an activated database as db2pd table statistics will be of little use otherwise; 
 # a first connection will activate an inactive database, so we have to check in a rather roundabout way;
 # first set a space-terminated variable for the database alias so a grep will not pick a false positive,
 # e.g., finding SAMPLE1 when it is looking for SAMPLE. 
 db=${dbname}" "
 typeset -i active=$(db2 "list active databases" | grep -c $db )
 if [[ $active -eq 1 ]]
 then
    : # no action
 else
    print ${BOLD} "$0 : database $dbname is not activated. Exiting ..." ${PLAIN}
    RESTORE_OUTPUT 
    exit 3
 fi
 
# Connection test - db2pd -tcbstats by itself does not require a connection, but catalog searches do.
 db2 -x connect to $dbname
 RC=$? ; DB2_RC_CHECK $RC 0 4 "$0 Connection error to database $dbname"

}

MAIN () {

 typeset -u dbname=$1 runstats_option=$2
 # See whether the caller wants to limit the statistics check to just some tablespaces;
 # if not, select all regular and large tablespaces with permanent data (no temporary or user temporary tablespaces).
 if [[ -s ${WORKDIR}/statschk_search_tbspaces ]]
 then
    # file is populated, copy to temp work directory
    cp -p ${WORKDIR}/statschk_search_tbspaces ${TMPDIR}/statschk_search_tbspaces.${TS} 
 else
    db2 -x "select tbspaceid 	         \
	    from   syscat.tablespaces	 \
	    where  datatype in ('A','L') \
	    order by tbspaceid for fetch only" > ${TMPDIR}/statschk_search_tbspaces.${TS}
 fi
 RC=$? ; DB2_RC_CHECK $RC 0 5 "$0 Query error on catalog view syscat.tablespaces. Detail follows."

 # The awkin scripts comment the logic for parsing db2pd output.
 > ${TMPDIR}/1.awk.out.${TS} 
 > ${TMPDIR}/2.awk.out.${TS} 
 for tbspaceid in `cat ${TMPDIR}/statschk_search_tbspaces.${TS}`
 do
     eval db2pd -tcbstats TbspaceID=${tbspaceid} -db ${dbname} -o ${TMPDIR}/db2pd.out.${tbspaceid}.${TS} 1> /dev/null # suppress informational messages
     # this return code check is of limited use, since db2pd still produces RC=0 for bad calls:
     RC=$? ; DB2_RC_CHECK $RC 0 6 "$0 call error on db2pd -tcbstats TbspaceID=${tbspaceid} -db ${dbname} -o ${TMPDIR}/db2pd.out.${tbspaceid}.${TS}"
     awk -f ${WORKDIR}/1.awkin ${TMPDIR}/db2pd.out.${tbspaceid}.${TS} >> ${TMPDIR}/1.awk.out.${TS}
     awk -f ${WORKDIR}/2.awkin ${TMPDIR}/db2pd.out.${tbspaceid}.${TS} >> ${TMPDIR}/2.awk.out.${TS}
 done

 # Sort both outputs on join field (table address)
 sort ${TMPDIR}/1.awk.out.${TS} > ${TMPDIR}/1.awk.out.sorted.${TS}
 sort ${TMPDIR}/2.awk.out.${TS} > ${TMPDIR}/2.awk.out.sorted.${TS}
 # Join on table address to obtain the initial list of fully qualified tables for RUNSTATS:
 join -o 1.2 1.3 -1 1  ${TMPDIR}/1.awk.out.sorted.${TS} ${TMPDIR}/2.awk.out.sorted.${TS} | sed 's/ /./' | sort -u > ${TMPDIR}/${logprefix}.tablist.${TS}
 
 # Add any tables that will always be RUNSTATS'D regardless
 if [[ -s ${WORKDIR}/statschk_include_tbspaces ]]
 then
    # file has size > 0
    for tbspaceid in `cat ${WORKDIR}/statschk_include_tbspaces`
    do
    db2 -x "select distinct rtrim(tabschema)||'.'||tabname \
            from   syscat.datapartitions                   \
            where  tbspaceid = $tbspaceid   		   \
            order  by 1 for fetch only" >> ${TMPDIR}/${logprefix}.tablist.${TS}
    RC=$? ; DB2_RC_CHECK $RC 0 7 "$0 Query error on catalog view syscat.datapartitions. Detail follows."
    done 
    # The search and include files could overlap, so remove duplicates:
    sort -u ${TMPDIR}/${logprefix}.tablist.${TS} > ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}
 else
    mv ${TMPDIR}/${logprefix}.tablist.${TS} ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}
 fi

 # RUNSTATS on identified tables, if requested:
 if [[ $runstats_option = "STATS" ]]
 then
    for name in `cat ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}`
    do
        schema=${name%%\.*}
        tabname=${name##*\.}
        RUNSTATS $schema $tabname
        RC=$?
        if [[ $RC -gt 0 ]]
        then
           print ${BOLD} "$0 : error in RUNSTATS() function, please review log above." ${PLAIN}
           RESTORE_OUTPUT 
           exit $RC
        fi
    done
 fi
}

FINAL () {
 db2 -v "terminate"
 # remove temporary work files from previous runs (those with a timestamp suffix not matching this run)
 find ${TMPDIR} \! -name "*.${TS}" -type f -exec rm -f {} \;
 print "$0 : log output is in ${LOG}"
 if [[ $option = "STATS" ]]
 then
    print "$0 : RUNSTATS was executed for tables in list $LOGDIR}/${logprefix}.runstats_tablelist.${TS}"
 else
    print "$0 : The list of tables to consider for RUNSTATS is in ${LOGDIR}/${logprefix}.runstats_tablelist.${TS}"
 fi
 print "$(date +"%d/%m/%Y %H:%M:%S") ===> $0 ends with maximum return code $RC."
 RESTORE OUTPUT 
}

INITIAL $1 $2
MAIN    $1 $2
FINAL   
